/* Program: ncoct.cc
   $Id: ncoct.cc,v 1.2 2007/03/01 11:08:02 raymond Exp $
   Author: Beverly Davis
   Modified by: David J. Raymond
   Copyright: (C) 2007 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Purpose: to read and write netCDF files in Octave
   Needed: Octave, and netCDF
   Compiled with: mkoctfile ncoct.cc -lnetcdf
   Usage: Called from Octave.  Does different operations depending on the 
      first argument given.  
         _operation_       _call in Octave_ 
         open a file       ncoct("open", fname)
         close a file      ncoct("close", ncid)
         create a file     ncoct("create", fname)
         add a dimension   ncoct("putdim", ncid, dname, ddata)
         read a variable   ncoct("getvar", ncid, vname)
         add a variable    ncoct("putvar", ncid, vname, rowid, colid, vdata)
      where
         fname:   netCDF file's name string
         ncid:    netCDF file id previously returned by open or create
         dname:   dimension name string
         ddata:   vector of dimension data
         vname:   variable field name string
         vdata:   matrix of variable field data
         rowid:   dimension id (previously returned by putdim) corresponding 
            to the rows of vdata
         colid:   dimension id (previously returned by putdim) corresponding 
            to the columns of vdata
    Notes: While this program is written in C++, it uses the "C interface" to 
       netCDF.  For more information about the netCDF C interface see "NetCDF 
       User's Guide for C", currently (as of March 31, 2004) available at 
       http://www.unidata.ucar.edu/packages/netcdf/guidec/
*/

#include <string>
#include <octave/oct.h>
#include <netcdf.h>

/* global history attribute for netCDF files created */
#define HIST "Created by Octave:\n"
/* absolute value of data range attribute for variables and dimensions */
#define DATARANGE 9.999e29
/* macro for dealing with netCDF errors */
#define CHECKNC if (ncstatus != NC_NOERR) {nc_enddef(ncid); error("%s line %d\n%s\n", __FILE__, __LINE__, nc_strerror(ncstatus)); return -1.0; }


using std::string;

/* functions dealing with the file */
double open(int nargin, octave_value_list args);
double close(int nargin, octave_value_list args);
double create(int nargin, octave_value_list args);

/* functions dealing with the dimensions */
double putdim(int nargin, octave_value_list args);

/* functions dealing with the variables */
octave_value getvar(int nargin, octave_value_list args);
double putvar(int nargin, octave_value_list args);


/* The main function to call in octave 
   It determines which operation is requested
   and calls the appropriate function */
DEFUN_DLD (ncoct, args, ,
  "Octave NetCDF interface.")
{
   /* if no arguments or non-string 1st argument
      give list of possible 1st arguments */
   int nargin = args.length();
   if (nargin == 0  || !args(0).is_string()){
      error("ncoct is a Octave netCDF interface.\nDifferent functions are used by giving different first arguments.\nThe functions currently available are: \n   \"open\"     open a file with read access\n   \"close\"    close a file\n   \"create\"   create a file with read and write access\n   \"putdim\"   add a dimension\n   \"putvar\"   add a variable field\n   \"getvar\"   read a variable field\n");
      return (octave_value) -1.0;
   }

   /* see if 1st argument corresponds to a function if so, run it */
   string fun = args(0).string_value();

   if (!strcmp(fun.c_str(), "open")){ 
      octave_value retval = (octave_value) open(nargin, args);
      /* return to Octave the netCDF id (>=0) of opened file as a double */
      return retval; 
   }

   if (!strcmp(fun.c_str(), "close")){
      octave_value retval = (octave_value) close(nargin, args);
      return retval;   /* return 0.0 if file properly closed */
   }

   if (!strcmp(fun.c_str(), "create")){
      octave_value retval = (octave_value) create(nargin, args);
      /* return to Octave the netCDF id (>=0) of created file as a double */
      return retval; 
   }

   if (!strcmp(fun.c_str(), "putdim")){
      octave_value retval = (octave_value) putdim(nargin, args);
      /* return to Octave the created dimension's id (>=0) as a double */
      return retval;
   }

   if (!strcmp(fun.c_str(), "getvar")){
      octave_value_list retval = getvar(nargin, args);
      /* return to Octave the variable's values as a matrix */
      return retval;
   }

   if (!strcmp(fun.c_str(), "putvar")){
      octave_value retval = (octave_value) putvar(nargin, args);
      /* return to Octave the created variable's id (>=0) as a double */
      return retval;
   }

   error("ncoct is a Octave netCDF interface.\nDifferent functions are used by giving different first arguments.\nThe functions currently available are: \n   \"open\"     open a file with read access\n   \"close\"    close a file\n   \"create\"   create a file with read and write access\n   \"putdim\"   add a dimension\n   \"putvar\"   add a variable field\n   \"getvar\"   read a variable field\n");
   return (octave_value) -1.0; 
}


/* Function to open file */
double open(int nargin, octave_value_list args){
   /* check arguments */
   if ( !(nargin==2 && args(1).is_string()) ){
      error("Usage: ncoct(\"open\", fname)\nopen a netCDF file for reading\nWhere\n   fname (a string) is the name of the file to be opened;\nReturns the netCDF id of the opened file if successful.\n");
      return -1.0;
   }
   string fname = args(1).string_value();
   /* open file */
   int ncid, ncstatus;
   ncstatus = nc_open(fname.c_str(), NC_NOWRITE, &ncid);
   CHECKNC;
   return (double) ncid;
}

/* Function to close file */
double close(int nargin, octave_value_list args){
   /* check arguments */
   if ( !(nargin==2 && args(1).is_scalar_type() && args(1).is_real_type()) ){
      error("Usage: ncoct(\"close\", ncid)\nclose a netCDF file\nWhere\n   ncid (a real scalar) is the netCDF id of the file to be closed.\nReturns 0.0 if successful.\n");
      return -1.0;
   }
   int ncid = (int) args(1).double_value();
   /* close file */
   int ncstatus;
   ncstatus = nc_close(ncid);
   CHECKNC;
   return (double) 0.0;
}

/* Function to create file */
double create(int nargin, octave_value_list args){
   /* check arguments */
   if ( !(nargin==2 && args(1).is_string()) ){
      error("Usage: ncoct(\"create\", fname)\ncreate a netCDf file for reading and writing\nWhere\n   fname (a string) is the name of the file to create.\nAny existing file with the same name is overwritten.  Returns the netCDF id of the created file if successful.\n");
      return -1.0;
   }
   string fname = args(1).string_value();

   /* create the file */
   int ncstatus, ncid; 
   ncstatus = nc_create(fname.c_str(), NC_CLOBBER, &ncid);
   CHECKNC;

   /* add a history attribute line */
   ncstatus = nc_put_att_text(ncid, NC_GLOBAL, "history", strlen(HIST), HIST);
   CHECKNC;

   /* take file out of define mode */
   ncstatus = nc_enddef(ncid);
   CHECKNC;

   return (double) ncid;
}


/* Function to add a dimension */
double putdim(int nargin, octave_value_list args){
   /* check arguments */
   if ( !(nargin==4 && args(1).is_scalar_type() && args(1).is_real_type() && args(2).is_string() && args(3).is_real_type() && (args(3).rows()==1 || args(3).columns()==1)) ){
      error("Usage: ncoct(\"putdim\", ncid, dname, ddata)\nadd a dimension\nWhere\n   ncid (a real scalar) is the netCDF id of the file to add the dimension to\n   dname (a string) is the name of the dimension to add\n   ddata (a real vector) is the vector containing the data for the dimension\nReturns the netCDF id of the added dimension if successful.\n");
      return -1.0;
   }
   int ncid = (int) args(1).double_value();
   string dname = args(2).string_value();
   ColumnVector ddata = (ColumnVector) args(3).vector_value();

   /* get size of vector to make into dimesnion */
   size_t dsize = (size_t) ddata.length();

   /* put file in define mode */
   int ncstatus;
   ncstatus = nc_redef(ncid);
   CHECKNC;

   /* add the dimension */
   int dimid;
   ncstatus = nc_def_dim(ncid, dname.c_str(), dsize, &dimid);
   CHECKNC;

   /* add a corresponding index field */
   int varid;
   ncstatus = nc_def_var(ncid, dname.c_str(), NC_DOUBLE, 1, &dimid, &varid);
   CHECKNC;

   /* add the valid range attribute to index field */
   double rangevals[] = {-DATARANGE, DATARANGE};
   ncstatus = nc_put_att_double(ncid, varid, "valid_range", NC_DOUBLE, (size_t)2, rangevals);
   CHECKNC;

   /* take file out of define mode */
   ncstatus = nc_enddef(ncid);
   CHECKNC;

   /* put the values into the C++ array */
   double darray[dsize];
   int i;
   for (i = 0; i < dsize; i++){
      darray[i] = ddata(i);
   }

   /* transfer the data */
   ncstatus = nc_put_var_double(ncid, varid, darray);
   CHECKNC;

   return (double) dimid;
}


/* Function to get all values of a variable */
octave_value getvar(int nargin, octave_value_list args){
   /* check arguments */
   if ( !(nargin==3 && args(1).is_scalar_type() && args(1).is_real_type() && args(2).is_string()) ){
      error("Usage: ncoct(\"getvar\", ncid, vname)\nread a variable\nWhere\n   ncid (a real scalar) is the netCDF id of the file to get the variable from\n   vname (a string) is the name of the varaible to get\nReturns the variable as an Octave matrix if successful.\n"); 
      return -1.0;
   }
   int ncid = (int) args(1).double_value();
   string vname = args(2).string_value();

   /* get variable id */
   int varid, ncstatus;
   ncstatus = nc_inq_varid(ncid, vname.c_str(), &varid);
   CHECKNC;

   /* get number of dimensions and dimension ids*/
   int ndims, dimids[NC_MAX_VAR_DIMS];
   ncstatus = nc_inq_var(ncid, varid, 0, 0, &ndims, dimids, 0);
   CHECKNC;

   /* check that there are no more than 2 dimensions */
   if (ndims > 2){  
      error("%s line %d\nOctave can't handle more than 2 dimensions.\n", __FILE__, __LINE__);
      return -1.0;
   }

   /* get current length of dimension/s */
   int idim, size[] = {1,1}; 
   size_t dimlen; 
   for (idim=0; idim < ndims; idim++){
      ncstatus = nc_inq_dimlen(ncid, dimids[idim], &dimlen);
      CHECKNC;
      size[idim] = (int) dimlen;
   }

   /* get the variable's values */
   double varvals[size[1]*size[0]];
   ncstatus = nc_get_var_double(ncid, varid, varvals);
   CHECKNC;

   /* put the values into the Octave matrix */
   int irow, icol;
   Matrix value(size[0],size[1]);
   for(irow=0;irow<size[0];irow++){
      for(icol=0; icol<size[1]; icol++){
         value(irow,icol)=varvals[icol+irow*size[1]];
      }
   }

   return value;
}


/* Function to add a variable */
double putvar(int nargin, octave_value_list args){
   /* check arguments */
   if ( !(nargin==6 && args(1).is_scalar_type() && args(1).is_real_type() && args(2).is_string() && args(3).is_scalar_type() && args(3).is_real_type() && args(4).is_scalar_type() && args(4).is_real_type() && args(5).is_real_type()) ){
      error("Usage: ncoct(\"putvar\", ncid, vname, rowid, colid, vdata)\nadd a variable\nWhere\n   ncid (a real scalar) is the netCDF id of the file to add the variable to\n   vname (a string) is the name of the varaible to add\n   rowid (a real scalar) is the netCDF id for the variable's rows dimension\n   colid (a real scalar) is the netCDF id for the variable's columns dimension\n   vdata (a real matrix) is the matrix containing the data for the variable\nIf vdata is a vector or a scalar, use a negative number to hold the place of the corresponding dimension id.\nReturns the netCDF id of the added variable if successful.\n");
      return -1.0;
   }
   int ncid = (int) args(1).double_value();
   string vname = args(2).string_value();
   int rcids[2];
   rcids[0] = (int) args(3).double_value();
   rcids[1] = (int) args(4).double_value();
   Matrix vdata = args(5).matrix_value();

   /* get size of variable */
   int size[2];
   size[0] = (int) vdata.rows();  /* number of rows */
   size[1] = (int) vdata.columns();  /* number of columns */

   /* check size of dimension/s */
   size_t dlen;
   int ncstatus, ndims=0, dimids[2];
   for (int i =0; i < 2; i++ ){
      if (rcids[i] >= 0){
         /* get size of dimension */
         ncstatus = nc_inq_dimlen(ncid, rcids[i], &dlen);
         CHECKNC;
         /* check that size of dimension and variable are the same */
         if (size[i] == dlen){
            /* add dimension id to list of dimension ids for nc_def_var */
            dimids[ndims] = rcids[i];
            /* and add 1 to the number of dimensions */
            ndims = ndims + 1;
         }
         else {
            error("%s line %d\ndimensions inconsistent with variable size and shape.\n", __FILE__, __LINE__);
            return -1;
         }
      }
      else {
         if (size[i]!=1){
            error("%s line %d\ndimensions inconsistent with variable size and shape.\n", __FILE__, __LINE__);
	    return -1;
         }
      }
   }

   /* put file in define mode */
   ncstatus = nc_redef(ncid);
   CHECKNC;

   /* create the variable */
   int varid;
   ncstatus = nc_def_var(ncid, vname.c_str(), NC_DOUBLE, ndims, dimids, &varid);
   CHECKNC;

   /* add the valid range attribute */
   double rangevals[] = {-DATARANGE, DATARANGE};
   ncstatus = nc_put_att_double(ncid, varid, "valid_range", NC_DOUBLE, (size_t)2, rangevals);
   CHECKNC;

   /* take file out of define mode */
   ncstatus = nc_enddef(ncid);
   CHECKNC;

   /* put the values into the C++ array */
   double varvals[size[0]*size[1]];
   int irow, icol;
   for(irow=0; irow<size[0]; irow++){
      for(icol=0; icol<size[1]; icol++){
	 varvals[icol+irow*size[1]] = vdata(irow,icol);
      }
   }

   /* give the netCDF variable values from the C++ array */ 
   ncstatus = nc_put_var_double(ncid, varid, varvals);
   CHECKNC;

   return (double) varid;
}

